;; Generated package description from isearch-mb.el  -*- no-byte-compile: t -*-
(define-package "isearch-mb" "0.7.0.20231020.185704" "Control isearch from the minibuffer" '((emacs "27.1")) :commit "8bc0d3fb7df5ffaaf787e0a1e7f5513067463e13" :authors '(("Augusto Stoffel" . "arstoffel@gmail.com")) :maintainer '("Augusto Stoffel" . "arstoffel@gmail.com") :keywords '("matching") :url "https://github.com/astoff/isearch-mb")
